/*
 *	Netgrind -- Saving to Files
 *
 *	(c) 2003 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#include "lib/lib.h"
#include "netgrind/pkt.h"
#include "netgrind/netgrind.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <netinet/in.h>

byte *save_dir;

void format_timestamp(byte *buf, u64 time)
{
  struct tm *tm;
  time_t t = time / 1000000;
  tm = localtime(&t);
  ASSERT(tm);
  int len = strftime(buf, TIMESTAMP_LEN, "%Y-%m-%d %H:%M:%S", tm);
  ASSERT(len);
  sprintf(buf+len, ".%03d", (uns)(time%1000000)/1000);
}

void sink_open(struct flow *f UNUSED, u64 when UNUSED)
{
}

void sink_close(struct flow *f UNUSED, int cause UNUSED, u64 when UNUSED)
{
}

void sink_input(struct flow *f UNUSED, int dir UNUSED, struct pkt *p)
{
  pkt_free(p);
}

struct appl_hooks appl_sink = {
  .open = sink_open,
  .input = sink_input,
  .close = sink_close
};

struct save_state {
  FILE *file[2];
};

static void save_open(struct flow *f, u64 when UNUSED)
{
  struct save_state *s = xmalloc(sizeof(*s));
  static uns save_counter;
  for (uns dir=0; dir<2; dir++)
    {
      byte name[256];
      sprintf(name, "%s/%06u-%d.%d.%d.%d:%d-%d.%d.%d.%d:%d-%c", save_dir, save_counter,
	      IPQUAD(f->saddr), ntohs(f->sport), IPQUAD(f->daddr), ntohs(f->dport),
	      'A' + dir);
      if (!(s->file[dir] = fopen(name, "w")))
	die("Unable to create %s: %m", name);
    }
  f->appl_data = s;
  save_counter++;
}

static void save_close(struct flow *f, int cause UNUSED, u64 when UNUSED)
{
  struct save_state *s = f->appl_data;
  fclose(s->file[0]);
  fclose(s->file[1]);
  xfree(s);
}

static void save_input(struct flow *f, int dir, struct pkt *p)
{
  struct save_state *s = f->appl_data;
  fwrite(p->data, pkt_len(p), 1, s->file[dir]);
  pkt_free(p);
}

struct appl_hooks appl_save = {
  .open = save_open,
  .input = save_input,
  .close = save_close
};

uns asave_width = 80;

struct asave_state {
  FILE *file;
  u64 start_time;
};

static void asave_event(struct asave_state *s, u64 time, byte *msg, ...)
{
  va_list args;

  va_start(args, msg);
  time -= s->start_time;
  fprintf(s->file, "%04d.%03d ", (int)(time/1000000), (int)(time%1000000)/1000);
  vfprintf(s->file, msg, args);
  va_end(args);
}

static void asave_open(struct flow *f, u64 when)
{
  struct asave_state *s = xmalloc(sizeof(*s));
  byte name[256], stamp[TIMESTAMP_LEN];
  static uns asave_counter;
  sprintf(name, "%s/%06u-%d.%d.%d.%d:%d-%d.%d.%d.%d:%d", save_dir, asave_counter++,
	  IPQUAD(f->saddr), ntohs(f->sport),
	  IPQUAD(f->daddr), ntohs(f->dport));
  if (!(s->file = fopen(name, "w")))
    die("Unable to create %s: %m", name);
  f->appl_data = s;
  s->start_time = when;
  asave_event(s, when, "Connection: %d.%d.%d.%d:%d -> %d.%d.%d.%d:%d\n",
	      IPQUAD(f->saddr), ntohs(f->sport), IPQUAD(f->daddr), ntohs(f->dport));
  format_timestamp(stamp, when);
  asave_event(s, when, "Initiated on: %s\n", stamp);
}

static void asave_close(struct flow *f, int cause, u64 when)
{
  struct asave_state *s = f->appl_data;
  asave_event(s, when, "Terminated: %s\n", flow_cause_names[cause]);
  asave_event(s, when, "TX %Ld bytes in %Ld packets, RX %Ld bytes in %Ld packets\n",
	      f->pipe[1].stat.bytes, f->pipe[1].stat.packets,
	      f->pipe[0].stat.bytes, f->pipe[0].stat.packets);
  asave_event(s, when, "Transferred %Ld bytes, TCP overhead %Ld bytes\n",
	      f->pipe[0].stat.bytes + f->pipe[1].stat.bytes,
	      f->stat_raw.bytes - (f->pipe[0].stat.bytes + f->pipe[1].stat.bytes));
  fclose(s->file);
  xfree(s);
}

static void asave_input(struct flow *f, int dir, struct pkt *p)
{
  struct asave_state *s = f->appl_data;

  if (!pkt_len(p))
    {
      asave_event(s, p->timestamp, "%s FIN\n", (dir ? ">>" : "<<"));
      return;
    }

  asave_event(s, p->timestamp, (dir ? ">>> " : "<<< "));
  uns len = pkt_len(p);
  uns cnt = 0;
  for (uns i=0; i<len; i++)
    {
      uns c = p->data[i];
      if (c == '\n')
	cnt = ~0U;
      else if (c >= 0x20 && c < 0x7f || c >= 0xa0)
	{
	  fputc(c, s->file);
	  cnt++;
	}
      else
	{
	  fprintf(s->file, "<%02x>", c);
	  cnt += 4;
	}
      if (cnt >= asave_width+10 || (cnt == asave_width && i < len-1 && p->data[i+1] != '\n'))
	{
	  if (c != '\n')
	    fputc('\\', s->file);
	  fputc('\n', s->file);
	  if (i < len-1)
	    asave_event(s, p->timestamp, (dir ? ">>> " : "<<< "));
	  cnt = 0;
	}
    }
  if (cnt)
    {
      fputc('\\', s->file);
      fputc('\n', s->file);
    }
  pkt_free(p);
}

struct appl_hooks appl_asave = {
  .open = asave_open,
  .input = asave_input,
  .close = asave_close
};

struct summary_state {
  u64 estab_time;
};

static void summary_open(struct flow *f, u64 when)
{
  struct summary_state *s = xmalloc(sizeof(*s));
  f->appl_data = s;
  s->estab_time = when;
}

static void summary_close(struct flow *f, int cause, u64 when)
{
  static int summary_cnt;
  if (!summary_cnt++)
    printf("# timestamp             source                destination              elapsed close   tx      rx  oh    KB/s\n");
         /* 2003-06-06 22:38:34.076 81.27.194.19:1165     195.113.31.123:22          1.604 OK    1991   12656  11%  8.914 */

  struct summary_state *s = f->appl_data;
  u64 duration = when - s->estab_time;
  byte src[22], dst[22];
  byte stamp[TIMESTAMP_LEN];
  format_timestamp(stamp, s->estab_time);
  sprintf(src, "%d.%d.%d.%d:%d", IPQUAD(f->saddr), ntohs(f->sport));
  sprintf(dst, "%d.%d.%d.%d:%d", IPQUAD(f->daddr), ntohs(f->dport));
  double braw = f->stat_raw.bytes;
  double beff = f->pipe[0].stat.bytes + f->pipe[1].stat.bytes;
  double over = 100 * ((braw - beff) / braw);
  double avs = duration ? ((beff / ((double)duration/1000000)) / 1024) : 1e30;
  printf("%s %-21s %-21s %6d.%03d %s %7Ld %7Ld %3d%% ",
	 stamp, src, dst,
	 (uns)(duration/1000000), (uns)(duration%1000000)/1000, flow_cause_names_short[cause],
	 f->pipe[1].stat.bytes, f->pipe[0].stat.bytes,
	 CLAMP((int)over,0,100));
  printf((avs < 100) ? "%6.3f" : "%6.0f", MIN(avs, 999999));
  putchar('\n');
  xfree(s);
}

struct appl_hooks appl_summary = {
  .open = summary_open,
  .input = sink_input,
  .close = summary_close
};
