/*
 *	Netgrind -- TCP Layer Analyser
 *
 *	(c) 2003 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#undef LOCAL_DEBUG

#include "lib/lib.h"
#include "lib/heap.h"
#include "netgrind/pkt.h"
#include "netgrind/netgrind.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

/*** TCP LAYER ***/

struct pkt_stats stat_tcp_in, stat_tcp_invalid, stat_tcp_badsum, stat_tcp_unmatched,
  stat_tcp_on_closed, stat_tcp_bad_state;
uns cnt_tcp_flows, cnt_tcp_causes[CAUSE_MAX];
uns tcp_arrival_times, tcp_wait_for_ack;

struct appl_hooks *tcp_default_appl;

byte *flow_state_names[] = { "IDLE", "SYNSENT", "SYNACK", "ESTAB", "FINSENT", "FINISH" };
byte *flow_cause_names[] = { "Close", "Reset", "Timeout", "Doomsday", "Corrupted" };
byte *flow_cause_names_short[] = { "OK", "RE", "TO", "DD", "CO" };

uns tcp_num_flows, tcp_max_flows;
static struct flow **flow_hash;
static struct flow **flow_heap;

static uns flow_calc_hash(u32 saddr, u32 daddr, u32 sport, u32 dport)
{
  saddr = (saddr >> 16) | (saddr << 16);
  daddr = (daddr >>  8) | (daddr << 24);
  sport <<= 7;
  dport <<= 21;
  return (saddr + daddr + sport + dport) % tcp_max_flows;
}

#define FLOW_HEAP_LESS(a,b) (a->timeout < b->timeout)
#define FLOW_HEAP_SWAP(h,a,b,t) do { t=h[a]; h[a]=h[b]; h[b]=t; h[a]->heap_pos=a; h[b]->heap_pos=b; } while(0)

static void flow_rehash(void)
{
  uns omax = tcp_max_flows;
  struct flow **ohash = flow_hash;

  if (flow_heap)
    xfree(flow_heap);
  if (tcp_max_flows)
    tcp_max_flows = nextprime(2*tcp_max_flows);
  else
    tcp_max_flows = 3;
  // DBG("Rehashing to %d buckets\n", tcp_max_flows);
  flow_hash = xmalloc_zero(sizeof(struct flow *) * tcp_max_flows);
  flow_heap = xmalloc_zero(sizeof(struct flow *) * (tcp_max_flows+1));
  tcp_num_flows = 0;
  for (uns i=0; i<omax; i++)
    {
      struct flow *f = ohash[i];
      while (f)
	{
	  struct flow *n = f->hash_next;
	  uns h = flow_calc_hash(f->saddr, f->daddr, f->sport, f->dport);
	  f->hash_next = flow_hash[h];
	  flow_hash[h] = f;
	  flow_heap[++tcp_num_flows] = f;
	  f->heap_pos = tcp_num_flows;
	  f = n;
	}
    }
  if (ohash)
    xfree(ohash);
  HEAP_INIT(struct flow *, flow_heap, tcp_num_flows, FLOW_HEAP_LESS, FLOW_HEAP_SWAP);
}

static struct flow *flow_lookup(u32 saddr, u32 daddr, u32 sport, u32 dport)
{
  uns h = flow_calc_hash(saddr, daddr, sport, dport);
  for (struct flow *f = flow_hash[h]; f; f=f->hash_next)
    if (f->saddr == saddr && f->daddr == daddr &&
	f->sport == sport && f->dport == dport)
      return f;
  return NULL;
}

static struct flow *flow_create(u32 saddr, u32 daddr, u32 sport, u32 dport)
{
  if (tcp_num_flows >= tcp_max_flows)
    flow_rehash();
  uns h = flow_calc_hash(saddr, daddr, sport, dport);
  struct flow *f = xmalloc_zero(sizeof(struct flow));
  f->saddr = saddr;
  f->daddr = daddr;
  f->sport = sport;
  f->dport = dport;
  f->timeout = ~0U;
  f->hash_next = flow_hash[h];
  flow_hash[h] = f;
  flow_heap[++tcp_num_flows] = f;
  f->heap_pos = tcp_num_flows;
  cnt_tcp_flows++;
  return f;
}

static void flow_set_timeout(struct flow *f, u32 when)
{
  f->timeout = when;
  HEAP_CHANGE(struct flow *, flow_heap, tcp_num_flows, FLOW_HEAP_LESS, FLOW_HEAP_SWAP, f->heap_pos);
}

static uns timestamp_to_now(u64 timestamp)
{
  return timestamp >> 20;
}

static u64 now_to_timestamp(uns now)
{
  return (u64)now << 20;
}

static inline int tcp_seq_le(u32 a, u32 b)
{
  return ((b - a) < 0x80000000);
}

static inline int tcp_seq_lt(u32 a, u32 b)
{
  return (a != b && tcp_seq_le(a, b));
}

static void tcp_time_step(uns now, uns doomsday)
{
  while (tcp_num_flows && (flow_heap[1]->timeout <= now || doomsday))
    {
      struct flow *f = flow_heap[1];
      HEAP_DELMIN(struct flow *, flow_heap, tcp_num_flows, FLOW_HEAP_LESS, FLOW_HEAP_SWAP);
      DBG("%s for flow %p(%s/%s) now=%d\n", (doomsday ? "DOOMSDAY" : "TIMEOUT"), f,
	  flow_state_names[f->pipe[0].state], flow_state_names[f->pipe[1].state], now);
      if (f->pipe[0].state != FLOW_FINISHED || f->pipe[1].state != FLOW_FINISHED)
	{
	  uns cause;
	  if (f->cnt_unexpected)
	    cause = CAUSE_CORRUPT;
	  else if (doomsday)
	    cause = CAUSE_DOOMSDAY;
	  else
	    cause = CAUSE_TIMEOUT;
	  cnt_tcp_causes[cause]++;
	  f->appl->close(f, cause, now_to_timestamp(now));
	}
      for (uns i=0; i<2; i++)
	pkt_flush_queue(&f->pipe[i].queue);
      uns h = flow_calc_hash(f->saddr, f->daddr, f->sport, f->dport);
      struct flow **gg = &flow_hash[h];
      for(;;)
	{
	  ASSERT(*gg);
	  if (*gg == f)
	    {
	      *gg = f->hash_next;
	      break;
	    }
	  gg = &(*gg)->hash_next;
	}
      xfree(f);
    }
}

static void tcp_enqueue_data(struct pipe *b, struct pkt *p)
{
  struct pkt *q, *prev, *new;
  u32 last_seq;

  DBG("DATA:");
  if (tcp_seq_lt(b->queue_start_seq, p->seq) && p->seq - b->queue_start_seq >= 0x40000)
    {
      DBG(" OUT OF WINDOW (q-start=%u)", b->queue_start_seq);
      pkt_free(p);
      return;
    }
  prev = (struct pkt *) &b->queue.head;
  last_seq = b->last_acked_seq;
  while (p)
    {
      if (tcp_seq_lt(p->seq, last_seq))
	{
	  if (tcp_seq_le(p->seq + pkt_len(p), last_seq))
	    {
	      DBG(" have");
	      pkt_free(p);
	      return;
	    }
	  pkt_pop(p, p->seq + pkt_len(p) - last_seq);
	  p->seq = last_seq;
	  DBG(" clip");
	}
      q = list_next(&b->queue, &prev->n);
      if (q && tcp_seq_le(q->seq, p->seq))
	{
	  /* next packet starts before us => skip it */
	  prev = q;
	  last_seq = q->seq + pkt_len(q);
	}
      else
	{
	  new = NULL;
	  if (q && tcp_seq_lt(q->seq, p->seq + pkt_len(p)))
	    {
	      /* overlap with next packet => split */
	      DBG(" split");
	      uns keeplen = q->seq - p->seq;
	      uns newlen = pkt_len(p) - keeplen;
	      new = pkt_new(0, newlen);
	      memcpy(pkt_append(new, newlen), pkt_unappend(p, newlen), newlen);
	      new->seq = p->seq + keeplen;
	    }
	  DBG(" insert");
	  list_insert(&p->n, &prev->n);
	  prev = p;
	  last_seq = p->seq + pkt_len(p);
	  p = new;
	}
    }
}

void tcp_got_packet(struct iphdr *iph, struct pkt *p)
{
  struct tcphdr *tcph;
  struct {
    u32 src;
    u32 dst;
    byte zero;
    byte proto;
    u16 len;
  } fakehdr;
  struct pkt *q;
  uns now = timestamp_to_now(p->timestamp);
  struct flow *f = NULL;

  tcp_time_step(now, 0);

  pkt_account(&stat_tcp_in, p);
  if (!(tcph = pkt_peek(p, sizeof(*tcph))))
    goto invalid;
  uns hdrlen = 4*tcph->doff;
  if (hdrlen < sizeof(*tcph) || hdrlen > pkt_len(p))
    goto invalid;
  fakehdr.src = iph->saddr;
  fakehdr.dst = iph->daddr;
  fakehdr.zero = 0;
  fakehdr.proto = IPPROTO_TCP;
  fakehdr.len = htons(pkt_len(p));
  uns sum = tcpip_calc_checksum(&fakehdr, sizeof(fakehdr), 0);
  sum = tcpip_calc_checksum(p->data, pkt_len(p), sum);
  if (!tcpip_verify_checksum(sum))
    {
      pkt_account(&stat_tcp_badsum, p);
      goto drop;
    }
  /* XXX: Check TCP options? */

  u32 seq = ntohl(tcph->seq);
  u32 ack = ntohl(tcph->ack_seq);
  DBG("TCP %08x %08x %04x %04x seq=%u len=%u end=%u ack=%u%s%s%s%s%s%s now=%d\n",
      ntohl(iph->saddr), ntohl(iph->daddr), ntohs(tcph->source), ntohs(tcph->dest),
      seq, pkt_len(p) - hdrlen, seq + pkt_len(p) - hdrlen, ack,
      (tcph->fin ? " FIN" : ""),
      (tcph->syn ? " SYN" : ""),
      (tcph->rst ? " RST" : ""),
      (tcph->psh ? " PSH" : ""),
      (tcph->ack ? " ACK" : ""),
      (tcph->urg ? " URG" : ""),
      now);

  struct pipe *a, *b;
  if (f = flow_lookup(iph->saddr, iph->daddr, tcph->source, tcph->dest))
    {
      a = &f->pipe[0];
      b = &f->pipe[1];
    }
  else if (f = flow_lookup(iph->daddr, iph->saddr, tcph->dest, tcph->source))
    {
      a = &f->pipe[1];
      b = &f->pipe[0];
    }
  else
    {
      /* Flow not found, if it's a SYN packet, go create it */
      if (tcph->syn && !tcph->ack && !tcph->rst && !tcph->fin)
	{
	  f = flow_create(iph->saddr, iph->daddr, tcph->source, tcph->dest);
	  f->appl = tcp_default_appl;
	  f->appl->open(f, p->timestamp);
	  a = &f->pipe[0];
	  b = &f->pipe[1];
	  list_init(&a->queue);
	  a->syn_or_fin_seq = a->last_acked_seq = seq;
	  a->state = FLOW_SYN_SENT;
	  list_init(&b->queue);
	  b->state = FLOW_IDLE;
	  DBG("\t%p NEW\n", f);
	  pkt_account(&f->stat_raw, p);
	  goto drop;
	}
      DBG("\tUnmatched\n");
      pkt_account(&stat_tcp_unmatched, p);
      goto drop;
    }
  pkt_account(&f->stat_raw, p);
  pkt_pop(p, hdrlen);

  DBG("\t%p %s (%s/%s) ", f, (a == &f->pipe[0] ? "A->B" : "B->A"), flow_state_names[f->pipe[0].state], flow_state_names[f->pipe[1].state]);
  if (a->state == FLOW_FINISHED && b->state == FLOW_FINISHED)
    {
      /* XXX: Here we probably should handle SYN (fast recycling of ports) */
      DBG("closed\n");
      pkt_account(&stat_tcp_on_closed, p);
      f = NULL;			     /* Avoid shifting flow timeout */
      goto drop;
    }

  if (tcph->rst)
    {
      DBG("RESET\n");
      cnt_tcp_causes[CAUSE_RESET]++;
      f->appl->close(f, CAUSE_RESET, p->timestamp);
      a->state = b->state = FLOW_FINISHED;
      goto drop;
    }

  if (tcph->syn)
    {
      if (tcph->fin || pkt_len(p))
	goto inval;
      if (tcph->ack)
	{			/* SYN ACK */
	  if (b->state == FLOW_SYN_SENT && b->syn_or_fin_seq+1 == ack)
	    {
	      DBG("SYN ACK\n");
	      a->last_acked_seq = ack;
	      a->syn_or_fin_seq = seq;
	      a->queue_start_seq = ack;
	      a->state = FLOW_SYN_SENT_ACK;
	      b->last_acked_seq = seq;
	      goto drop;
	    }
	  else if (b->state == FLOW_ESTABLISHED)
	    goto dup;
	  else
	    goto unex;
	}
      else
	goto dup; /* otherwise SYN on already existing connection gets ignored */
    }

  if (tcph->ack)
    {
      if (tcp_seq_le(ack, a->last_acked_seq))
	DBG("DUP ACK, ");
      else
	{
	  a->last_acked_seq = ack;
	  if (tcp_wait_for_ack)
	    {
	      while ((q = list_head(&a->queue)) && tcp_seq_le(q->seq+pkt_len(q), ack))
		{
		  list_remove(&q->n);
		  a->queue_start_seq = q->seq + pkt_len(q);
		  if (!tcp_arrival_times)
		    q->timestamp = p->timestamp;
		  DBG("data(%Ld-%Ld), ", a->stat.bytes, a->stat.bytes+pkt_len(q)-1);
		  pkt_account(&a->stat, q);
		  f->appl->input(f, (a == &f->pipe[0]), q);
		}
	    }
	  if (b->state == FLOW_SYN_SENT_ACK && b->syn_or_fin_seq+1 == ack)
	    {
	      a->state = b->state = FLOW_ESTABLISHED;
	      a->queue_start_seq = ack;
	      DBG("ACKED SYN, ");
	    }
	  else if (b->state == FLOW_FIN_SENT && b->syn_or_fin_seq+1 == ack)
	    {
	      b->state = FLOW_FINISHED;
	      q = pkt_new(0, 0);
	      q->timestamp = p->timestamp;
	      f->appl->input(f, (b == &f->pipe[0]), q);
	      if (a->state == FLOW_FINISHED)
		{
		  DBG("CLOSED BOTH WAYS\n");
		  cnt_tcp_causes[CAUSE_CLOSE]++;
		  f->appl->close(f, CAUSE_CLOSE, p->timestamp);
		  goto drop;
		}
	      else
		DBG("CLOSED ONE-WAY, ");
	    }
	  else if (tcp_seq_lt(a->queue_start_seq, ack))
	    {
	      DBG("DAMNED, ACK FOR UNCAUGHT DATA!\n");
	      goto unexpected;
	    }
	  else if (b->state == FLOW_SYN_SENT_ACK || b->state == FLOW_SYN_SENT)
	    goto unex;
	}
    }

  if (tcph->fin)
    {
      if (a->state == FLOW_ESTABLISHED)
	{
	  a->state = FLOW_FIN_SENT;
	  a->syn_or_fin_seq = seq + pkt_len(p);
	  DBG("FIN SENT, waiting for FIN ACK, ");
	}
      else if (a->state == FLOW_FIN_SENT)
	;
      else
	goto unex;
    }

  if (!pkt_len(p))
    {
      DBG("EMPTY\n");
      goto drop;
    }

  if (b->state == FLOW_ESTABLISHED || b->state == FLOW_FIN_SENT || b->state == FLOW_FINISHED)
    {
      u64 arrival = p->timestamp;
      p->seq = seq;
      tcp_enqueue_data(b, p);
      if (!tcp_wait_for_ack)
	{
	  while ((q = list_head(&b->queue)) && q->seq == b->queue_start_seq)
	    {
	      list_remove(&q->n);
	      if (!tcp_arrival_times)
		q->timestamp = arrival;
	      DBG(", data(%Ld-%Ld)", b->stat.bytes, b->stat.bytes+pkt_len(q)-1);
	      pkt_account(&b->stat, q);
	      b->queue_start_seq += pkt_len(q);
	      f->appl->input(f, (b == &f->pipe[1]), q);
	    }
	}
      DBG("\n");
      goto done;
    }
  else
    goto unex;
  /* Not reached */

 dup:
  DBG("DUP\n");
 drop:
  pkt_free(p);
 done:
  if (f)
    {
      int timeout = 60;			/* Connection setup timeout */
      if (f->pipe[0].state == FLOW_FINISHED && f->pipe[1].state == FLOW_FINISHED)
	timeout = 30;			/* After connection close; later the packets are just unmatched */
      else if ((f->pipe[0].state == FLOW_ESTABLISHED || f->pipe[0].state == FLOW_FIN_SENT || f->pipe[0].state == FLOW_FINISHED) &&
	       (f->pipe[1].state == FLOW_ESTABLISHED || f->pipe[1].state == FLOW_FIN_SENT || f->pipe[1].state == FLOW_FINISHED))
	timeout = 900;			/* Data phase timeout */
      DBG("timeout at %d\n", now+timeout);
      flow_set_timeout(f, now + timeout);
    }
  return;

 unex:
  DBG("UNEXPECTED\n");
 unexpected:
  f->cnt_unexpected++;
  pkt_account(&stat_tcp_bad_state, p);
  goto drop;

 inval:
  DBG("???\n");
 invalid:
  pkt_account(&stat_tcp_invalid, p);
  goto drop;
}

void tcp_init(void)
{
  flow_rehash();
}

void tcp_cleanup(u64 timestamp)
{
  tcp_time_step(timestamp_to_now(timestamp), 1);
}
